import numpy as np 
from dataclasses import dataclass 


# General problem setup
@dataclass
class Problem:
    # workspace 
    # sim 
    # ws_x = [-3.0, 3.0]              # m 
    # ws_y = [-3.0, 3.0]              # m
    ws_x = [-4.0, 5.0]              # m 
    ws_y = [-3.0, 3.0]              # m
    # real experiment
    # ws_x = [-3.2, 3.2]            # m 
    # ws_y = [-1.2, 1.2]            # m

    # robot control bound
    robot_maxAcc = 1.0              # m/s^2
    robot_minAcc = -3.5
    robot_maxOmegaAcc = 1.0         # rad/s^2
    robot_minOmegaAcc = -3.5
    max_F_push = 100.0              # N

    # robot/object size, start and goal
    rob_size = 0.27                 # m (radius of the cylinder)
    rect_rob_size = [0.97, 0.67]    # m (W*L) on the x, y direction with regard to the robot frame
    # obj_size = [0.565, 0.755]       # m (W*L) on the x, y direction with regard to the object frame 75.5, 57
    obj_size = [0.345, 0.49]        # m (W*L) on the x, y direction with regard to the object frame
    obj_mass = 1.5     # kg
    obj_I_or = 1/12 * obj_mass * \
                    (obj_size[0]**2 + obj_size[1]**2)   # moment of inertia (rotation center is 
                                                        # the center of the object) (for rectangle)
    mu_g = 0.4                     # friction coefficient of the ground-object contact surface
    mu_p = 0.0                     # friction coefficient of the robot-object contact surface
    grav_acc = 9.8                 # gravity acceleration

    # bound
    robot_minVel = 0.01
    robot_maxVel = 0.40            # m/s
    robot_maxOmega = 0.3           # rad/s 
    d_max = obj_size[1]/2
    phi_max = np.pi
    d_maxVel = 0.5
    phi_maxVel = 0.5

    # system target
    obj_pos_goal = [0., -1.]        # m
    obj_theta_goal = [np.deg2rad(0.0)]# rad
    
    # General planning settings
    dt = 0.1                        # sampling time, s
    N = 30                          # horizon length
    T = N*dt                        # horizon time 
    nx = 9                          # state dimension 
    nx_dae = 11
    nx_rob = 5
    nu = 2                          # control dimension
    nz = 0                          # dimension of the algebraic variables
    nh = 0                          # number of inequality constraint functions
    nparam = 10                      # parameter dimension
    
    # MPC cost terms weights, can be real time param
    w_navi_error = 10.
    w_d_error = 0.
    w_phi_error = 0.

    w_px_error = 1.0
    w_py_error = 1.0
    w_theta_error = 0.0
    w_vel = 0.1
    w_omega = 0.1
    w_acc = 0.01
    w_omega_acc = 0.01
    w_f_push = 0.01

    # Feedback control gains
    k_rho = 1.0
    k_alpha = 8.0
    k_phi = -1.5

    # for dubin's car, push constraint
    v_omega_k = 0.2
    v_omega_k_max = 50.35  # maximum radius of turning left (left or right here should be comfirmed later)
    v_omega_k_min = -50.35  # maximum radius of turning right (should also be a positive value)
   

# General vector index
@dataclass
class Index:
    # control vector, u = [acc, omega_acc]
    u_acc_r = np.s_[0]              # 0
    u_omega_acc_r = np.s_[1]        # 1

    # state vector, q = [x_r, y_r, theta_r, v_r, d_theta_r, fai, d, d_fai, d_d]
    rob_state = np.s_[0:5]          # 0, 1, 2, 3, 4
    obj_state = np.s_[5:9]          # 5, 6, 7, 8 

    x_r = np.s_[0]                  # 0
    y_r = np.s_[1]                  # 1
    theta_r = np.s_[2]              # 2
    v_r = np.s_[3]                  # 3
    d_theta_r = np.s_[4]            # 4
    phi = np.s_[5]                  # 5
    d = np.s_[6]                    # 6
    d_phi = np.s_[7]                # 7
    d_d = np.s_[8]                  # 8

    # state vector for dae, x = [robot_pos_w_x, robot_pos_w_y, theta_r, v_r, d_theta_r, obj_pos_w_x, obj_pos_w_y, theta_o, vx_o, vy_o, d_theta_o]
    robot_pos_w_x = np.s_[0]        # 0
    robot_pos_w_y = np.s_[1]        # 1
    theta_r = np.s_[2]              # 2
    v_r = np.s_[3]                  # 3
    d_theta_r = np.s_[4]            # 4
    obj_pos_w_x = np.s_[5]          # 5
    obj_pos_w_y = np.s_[6]          # 6
    theta_o = np.s_[7]              # 7
    vx_o = np.s_[8]                 # 8
    vy_o = np.s_[9]                 # 9
    d_theta_o = np.s_[10]           # 10

    # param vector, customized for different formulation
    p_goal_pos_x = np.s_[0]
    p_goal_pos_y = np.s_[1]
    p_goal_theta = np.s_[2]
    # p_initial_state = np.s_[3:12]
    p_initial_state_robot = np.s_[3:8]
    p_v_omega_k_max = np.s_[8]   # max turning radius
    p_v_omega_k_min = np.s_[9]   # min turning radius
    p_R_pos_o_x = np.s_[10]      # object position in the robot frame (x)
    p_R_pos_o_y = np.s_[11]      # object position in the robot frame (y)

@dataclass
class Mpc_model:
    # MPC
    dt = 0             # sampling time
    N  = 0             # horizon length
    nx = 0             # number of equality constraints, x
    nu = 0             # number of control inputs, u
    ns = 0             # number of slacks, s
    np = 0             # number of real-time params on each stage, 
    # bound 
    xl = []
    xu = []
    ul = []
    uu = []
    sl = []
    su = []