#!/usr/bin/env python
import numpy as np
import rospy
import gym
import myenv
from timeit import default_timer as timer
from scipy.spatial.transform import Rotation as R

from geometry_msgs.msg import TwistStamped, Twist, Pose
from nav_msgs.msg import Odometry
from std_msgs.msg import Bool, Float64MultiArray

# from MPPI_vel import MPPI
from State_Estimator import State_Estimator
import sys
from controllers.mpc.stable_push_mpc_controller import Stable_push_mpc_controller
from general_problem_setup import Problem, Index
import utils.utils as utils
from myenv.stable_push_with_contact_switching_vis import Stable_push_with_contact_switching_vis

import gym
from stable_baselines3 import PPO
from stable_baselines3.a2c.a2c import A2C
from collections import OrderedDict

import sys
from pathlib import Path

FILE_THIS = Path(__file__).resolve()
PARENT = FILE_THIS.parent
GPARENT = FILE_THIS.parents[1]

class Husky_Control:
    def __init__(self):
        self.pr = Problem()
        self.index = Index()

        # ros publisher for control
        self.pub_husky_control = rospy.Publisher("/husky_velocity_controller/cmd_vel", Twist, queue_size=1) 

        # publisher for visualization
        self.pub_robot_mpc_x_plan = rospy.Publisher('/husky/visualization/mpc_x_plan', Float64MultiArray, queue_size=1)

        # subscribe
        self.topic_name_robot = "/Bebop1/position_velocity_orientation_estimation"
        self.topic_name_object = "/Bebop2/position_velocity_orientation_estimation"
        self.sub_robot_state_gazebo = rospy.Subscriber(self.topic_name_robot,Odometry, self.receive_robot_state)
        self.sub_object_state = rospy.Subscriber(self.topic_name_object, Odometry, self.receive_object_state)
        self.topic_ready_to_push = "/topic_ready_to_push"
        self.sub_ready_to_push = rospy.Subscriber(self.topic_ready_to_push , Bool, self.receive_ready_to_push_command)

        # Initializing robot mpc casadi
        self.mpc_controller = Stable_push_mpc_controller()
        recompile = True
        solver_build_name = 'Stable_push_mpc_controller_casadi'    
        self.mpc_controller.generate_solver(recompile, solver_build_name)

        self.obj_goal = self.pr.obj_pos_goal

        # init robot and object state
        self.robot_state =          [0.0, 0.0, 0.0, 0.0, 0.0]
        self.object_state =         [0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.full_state =           [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]
        self.prev_full_state =      [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def get_null_twist(self):
        cmd = Twist()
        cmd.linear.x = 0.0
        cmd.linear.y = 0.0
        cmd.linear.z = 0.0
        cmd.angular.x = 0.0
        cmd.angular.y = 0.0
        cmd.angular.z = 0.0
        return cmd

    def pub_control_action(self, robot_control_current):
        cmd = self.get_null_twist()
        cmd.linear.x = robot_control_current[0]
        cmd.angular.z = robot_control_current[1]
        self.pub_husky_control.publish(cmd)

    def pub_mpc_plan(self, mpc_path):
        all_x_coord = []
        all_y_coord = []
        all_x_coord_object = []
        all_y_coord_object = []

        for i in range(0, mpc_path.shape[1]):
            x_coord = mpc_path[:, i, 0]  # a list of x in one episode
            y_coord = mpc_path[:, i, 1]  # a list of y in one episode
            all_x_coord.append(x_coord)
            all_y_coord.append(y_coord)

        mpc_x_plan = np.concatenate((np.array([all_x_coord[0]]).T, np.array([all_y_coord[0]]).T), axis=1)

        # change to the world coordinate
        # robot_mpc_x_plan = mpc_x_plan + np.tile(self.pr.obj_pos_goal,(mpc_x_plan.shape[0],1))  # nz x N array
        robot_mpc_x_plan = mpc_x_plan
        robot_mpc_x_plan_msg = Float64MultiArray()
        robot_mpc_x_plan_msg.data = robot_mpc_x_plan.reshape((-1, 1), order='F')  # into a column vector
        self.pub_robot_mpc_x_plan.publish(robot_mpc_x_plan_msg)

    def receive_robot_state(self, odom_robot):

        r = R.from_quat([odom_robot.pose.pose.orientation.x, 
                            odom_robot.pose.pose.orientation.y, 
                            odom_robot.pose.pose.orientation.z, 
                            odom_robot.pose.pose.orientation.w])

        ypr = r.as_euler('zyx', degrees=False)[0]

        vel = np.sqrt(odom_robot.twist.twist.linear.x**2 + odom_robot.twist.twist.linear.y**2)
        omega = odom_robot.twist.twist.angular.z
        self.robot_state = [odom_robot.pose.pose.position.x, 
                                   odom_robot.pose.pose.position.y, ypr, vel, omega]

    def receive_object_state(self, obj_pos_msg):

        r = R.from_quat([obj_pos_msg.pose.pose.orientation.x, 
                            obj_pos_msg.pose.pose.orientation.y, 
                            obj_pos_msg.pose.pose.orientation.z, 
                            obj_pos_msg.pose.pose.orientation.w])

        ypr = r.as_euler('zyx', degrees=False)[0]

        vel_x = obj_pos_msg.twist.twist.linear.x
        vel_y = obj_pos_msg.twist.twist.linear.y
        omega = obj_pos_msg.twist.twist.angular.z
        self.object_state = [obj_pos_msg.pose.pose.position.x, obj_pos_msg.pose.pose.position.y, ypr, vel_x, vel_y, omega]

    def receive_ready_to_push_command(self, ready_to_push_command):
        self.ready_to_push = ready_to_push_command

    def check_for_anomalies(self, full_state):
        # check robot velocity
        full_state_new = full_state
        v_r = full_state[3]
        omega_r = full_state[4]

        if v_r > 2.5 or v_r < -2.5:
            full_state_new[3] = self.prev_full_state[3]
        if omega_r > 2.5 or omega_r < -2.5:
            full_state_new[4] = self.prev_full_state[4]

        # check object velocity
        v_xo = full_state[8]
        v_yo = full_state[9]
        omega_o = full_state[10]

        if v_xo > 2.5 or v_xo < -2.5:
            full_state_new[8] = self.prev_full_state[8]
        if v_yo > 2.5 or v_yo < -2.5:
            full_state_new[9] = self.prev_full_state[9]
        if omega_o > 2.5 or omega_o < -2.5:
            full_state_new[10] = self.prev_full_state[10]

        if full_state_new.any() != full_state.any():
            print("Weird velocities detected!")
        return full_state_new

    def get_state(self):
        # first set the previous state to remember it
        self.prev_full_state = self.full_state

        # update the current state
        full_state = self.robot_state + self.object_state
        full_state_arr = np.array(full_state)

        # check if there are some weird data values 
        full_state_arr = self.check_for_anomalies(full_state_arr)

        self.full_state = full_state_arr

        return full_state_arr

def run_feedback_obs_free_navi(husky_control, robot_pose_current, robot_pos_goal):
    # feedback control law 
    robot_pos_current = robot_pose_current[0:2]
    robot_theta_current = robot_pose_current[2]
    px_error = robot_pos_goal[0] - robot_pos_current[0]
    py_error = robot_pos_goal[1] - robot_pos_current[1]
    rho = np.sqrt(px_error ** 2 + py_error ** 2)
    alpha = -robot_theta_current + np.arctan2(py_error, px_error)
    phi = -robot_theta_current - alpha
    vel = husky_control.pr.k_rho * rho
    omega = husky_control.pr.k_alpha * alpha + husky_control.pr.k_phi * phi
    factor = min(1.0, husky_control.pr.robot_maxVel / vel)
    robot_control_current = [vel * factor, omega * factor]

    return robot_control_current

class Contact_switching_policy:
    def __init__(self, husky_control):
        self.env = gym.make('rob_navi-v2')
        self.env.init_path_candidates_both(1)
        self.model = PPO.load("/home/yujie/robot_pushing/stable_pushing_with_contact_switching/catkin_ws/src/PPO_switching_10311454.zip", env=self.env)
        self.obs = self.env.reset()

        self.W_o = husky_control.pr.rect_rob_size[0]
        self.L_o = husky_control.pr.rect_rob_size[1]
        self.W_r = husky_control.pr.rect_rob_size[0]
        self.L_r = husky_control.pr.rect_rob_size[1]

        self.ready_to_push = False

    def contact_switching_choice(self, robot_state_inGoalRefF, object_state_inGoalRefF, contact_state):
        obs = OrderedDict([
            ('robot_state', robot_state_inGoalRefF),
            ('contact_state', contact_state)
        ])        

        self.env.set_state(robot_state_inGoalRefF, object_state_inGoalRefF, contact_state)
        self.env.render()
        action, _state = self.model.predict(obs, deterministic=True)  # TODO: obs will be updated outside of the env, env is only for visualization

        self.env.show_paths_candidates_robot(self.env.mpc_controller.system_state_plan)

        # obs, reward, done, info = self.env.step(action_1)
        # action_2, _state = self.model.predict(obs, deterministic=True)
        # obs, reward, done, info = self.env.step(action_2)
        # action_3, _state = self.model.predict(obs, deterministic=True)

        # there are cases turn right three times, which can be simplified as turn left once
        # action_m_list = [action_1-1, action_2-1, action_3-1]

        # if abs(sum(action_m_list)) == 3:
        #     action = -(action_1-1)+1
        # elif abs(sum(action_m_list)) == 2:
        #     action = 2
        # else:
        #     action = action_1

        # contact state update
        contact_state_new = contact_state - (action-1)
        if contact_state_new == -1:
            contact_state_new = 3
        elif contact_state_new == -2:
            contact_state_new = 2
        elif contact_state_new == 4:
            contact_state_new = 0
        elif contact_state_new == 5:
            contact_state_new = 1  

        # if action != 1, calculate next contact point in the Goal frame
        if action != 1:
            # TODO decide which is the real next subgoal
            list_R_pose_o = np.array(  [[(self.L_o + self.W_r)/2, 0, np.pi/2],    # 'contact_state_0'
                                        [(self.W_o + self.W_r)/2, 0, 0],          # 'contact_state_1'
                                        [(self.L_o + self.W_r)/2, 0, -np.pi/2],   # 'contact_state_2'
                                        [(self.W_o + self.W_r)/2, 0, np.pi]    ]) # 'contact_state_3'
            R_pos_o = list_R_pose_o[contact_state_new]
            next_contact_point_inGoalRefF = np.array([0.0, 0.0, 0.0])
            next_contact_point_inGoalRefF[2] = robot_state_inGoalRefF[2] - (action-1)*np.pi/2
            angle = next_contact_point_inGoalRefF[2]%(2*np.pi)
            next_contact_point_inGoalRefF[2] = angle if angle < np.pi else angle - 2*np.pi
            next_contact_point_inGoalRefF[0:2] = object_state_inGoalRefF[0:2] - np.dot(utils.rot_M(next_contact_point_inGoalRefF[2]), np.array(R_pos_o[0:2]))

            self.dist = 0.3
            list_R_pose_o_withDist = np.array( [[(self.L_o + self.W_r)/2 + self.dist, 0, np.pi/2],    # 'contact_state_0'
                                                [(self.W_o + self.W_r)/2 + self.dist, 0, 0],          # 'contact_state_1'
                                                [(self.L_o + self.W_r)/2 + self.dist, 0, -np.pi/2],   # 'contact_state_2'
                                                [(self.W_o + self.W_r)/2 + self.dist, 0, np.pi]    ]) # 'contact_state_3'
            R_pos_o_withDist = list_R_pose_o_withDist[contact_state_new]
            rob_sub_goal_inGoalRefF = np.array([0.0, 0.0, 0.0])
            rob_sub_goal_inGoalRefF[2] = rob_sub_goal_inGoalRefF[2] - (action-1)*np.pi/2
            angle = rob_sub_goal_inGoalRefF[2]%(2*np.pi)
            rob_sub_goal_inGoalRefF[2] = angle if angle < np.pi else angle - 2*np.pi
            rob_sub_goal_inGoalRefF[0:2] = object_state_inGoalRefF[0:2] - np.dot(utils.rot_M(rob_sub_goal_inGoalRefF[2]), np.array(R_pos_o_withDist[0:2]))
        else:
            next_contact_point_inGoalRefF = robot_state_inGoalRefF[0:3]
            rob_sub_goal_inGoalRefF = robot_state_inGoalRefF[0:3]
        return action, next_contact_point_inGoalRefF, rob_sub_goal_inGoalRefF
        
def run_control():
    print("Starting Control...")

    # Initialize a ros node
    rospy.init_node('mpc_control', anonymous=False)

    # Create a husky robot object
    husky = Husky_Control()
    # contact_switching_policy = Contact_switching_policy(husky)

    # Create state estimation object 
    # mode = rospy.get_param('mode') # simulation or mocap
    mode = "mocap"
    # state_estimator = State_Estimator(mode)
    rospy.sleep(2)

    # Get state from the state estimator
    full_state = husky.get_state()
    # get current contact state
    # get states in the goal reference frame
    robot_state_inGoalRefF = full_state[0:5]
    robot_state_inGoalRefF[0:2] = robot_state_inGoalRefF[0:2] - np.array(husky.pr.obj_pos_goal)
    object_state_inGoalRefF = full_state[5:11]
    object_state_inGoalRefF[0:2] = object_state_inGoalRefF[0:2] - np.array(husky.pr.obj_pos_goal)
    O_p_r = np.dot(utils.rot_M(-full_state[7]), (full_state[0:2] - full_state[5:7])) # O_x_r = R^T(theta_o)*(W_x_r - W_x_o)
    if O_p_r[0] < 0 and abs(O_p_r[0]) > husky.pr.obj_size[0]/2.:
        contact_state = 1
    elif O_p_r[1] < 0 and -O_p_r[1] > -husky.pr.obj_size[1]/2.:
        contact_state = 0
    elif O_p_r[1] > 0 and O_p_r[1] > husky.pr.obj_size[1]/2.:
        contact_state = 2

    # Specify the control node frequency
    hz = 10
    rate = rospy.Rate(hz)
    rospy.sleep(0.5)

    disGoal = np.linalg.norm(full_state[5:7] - husky.pr.obj_pos_goal)

    while disGoal>0.15 :
        disGoal = np.linalg.norm(full_state[5:7] - husky.pr.obj_pos_goal)

        # Get state from the state estimator
        full_state = husky.get_state()
        robot_state = full_state[0:5]
        object_state = full_state[5:11]

        # get states in the goal reference frame
        # robot_state_inGoalRefF = full_state[0:5]
        # robot_state_inGoalRefF[0:2] = robot_state_inGoalRefF[0:2] - np.array(husky.pr.obj_pos_goal)
        # object_state_inGoalRefF = full_state[5:11]
        # object_state_inGoalRefF[0:2] = object_state_inGoalRefF[0:2] - np.array(husky.pr.obj_pos_goal)
        # O_p_r = np.dot(utils.rot_M(-full_state[7]), (full_state[0:2] - full_state[5:7]))  # O_x_r = R^T(theta_o)*(W_x_r - W_x_o)
        # if O_p_r[0] < 0 and abs(O_p_r[0]) > husky.pr.obj_size[0]/2.:
        #     contact_state = 1
        # elif O_p_r[1] < 0 and -O_p_r[1] > -husky.pr.obj_size[1]/2.:
        #     contact_state = 0
        # elif O_p_r[1] > 0 and O_p_r[1] > husky.pr.obj_size[1]/2.:
        #     contact_state = 2

        # check whether to switch contact 
        # switch_action, next_contact_point_inGoalRefF, rob_sub_goal_inGoalRefF = contact_switching_policy.contact_switching_choice(
        #                                         robot_state_inGoalRefF, object_state_inGoalRefF, contact_state
        #                                         )
        switch_action = 1

        if switch_action != 1:
            # human takeover the reposition
            # rostopic pub /topic_ready_to_push std_msgs/Bool True
            while not contact_switching_policy.ready_to_push:
                rospy.sleep(0.1)
            contact_switching_policy.ready_to_push = False
 
        else: # move without contact switch
            # Set state to the controller
            husky.mpc_controller.set_system_state_current(full_state[0:5])
            # Set goal position
            husky.mpc_controller.set_obj_pos_theta_goal(husky.obj_goal, 0)
            # run mpc
            start = timer()
            husky.mpc_controller.run_mpc(husky.pr.v_omega_k_max, husky.pr.v_omega_k_min, [0,0])
            end = timer()
            # print("Control Processing time: " + str(end - start))
            # retrieve info
            mpc_feasible = husky.mpc_controller.mpc_feasible
            action = husky.mpc_controller.robot_control_current
            # pub control signal
            husky.pub_control_action(action)
            print("Vel action: ", action)

            # pub robot planned trajectory
            husky.pub_mpc_plan(husky.mpc_controller.system_state_plan)

            # Get state from the state estimator
            full_state = husky.get_state()
            robot_state = full_state[0:5]
            object_state = full_state[5:11]
            
        
        # Sleep until 0.1s done
        rate.sleep()
        disGoal = np.linalg.norm(full_state[5:7] - husky.pr.obj_pos_goal)

if __name__ == '__main__':
    rospy.sleep(1.0)

    run_control()
