#!/usr/bin/env python

import numpy as np
import rospy
from std_msgs.msg import Float64MultiArray, Bool
from nav_msgs.msg import Odometry
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from scipy.spatial.transform import Rotation as R


import sys
from pathlib import Path

FILE_THIS = Path(__file__).resolve()
PARENT = FILE_THIS.parent
GPARENT = FILE_THIS.parents[1]
sys.path.append(str(PARENT))
sys.path.append(str(GPARENT))

from general_problem_setup import Problem, Index
import utils.utils as utils

class Husky_Visualization:
    def __init__(self):
        # Loading problem setup and indexing
        self.pr = Problem()
        self.index = Index()

        # Visualization flag
        self.sim_reset = False
        self.visible_robot_his_path = True
        self.visible_box_his_path = True

        # Visualization info
        self.robot_state = np.zeros(5)
        self.object_state = np.zeros(6)
        self.obj_pos_goal = self.pr.obj_pos_goal
        self.obj_theta_goal = self.pr.obj_theta_goal
        self.robot_mpc_plan = np.tile(np.zeros(self.pr.nx).reshape((-1, 1)), (1, self.pr.N))
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))

        self.W_r = self.pr.rect_rob_size[0]
        self.L_r = self.pr.rect_rob_size[1]
        self.W_o = self.pr.obj_size[0]
        self.L_o = self.pr.obj_size[1]

        # ROS subscribers
        self.topic_name_robot = "/Bebop1/position_velocity_orientation_estimation"
        self.topic_name_object = "/Bebop3/position_velocity_orientation_estimation"
        self.sub_robot_state_gazebo = rospy.Subscriber(self.topic_name_robot,Odometry, self.receive_robot_state)
        self.sub_object_state = rospy.Subscriber(self.topic_name_object, Odometry, self.receive_object_state)
        self.sub_sim_reset = rospy.Subscriber('/husky/visualization/reset', Bool, self.receive_sim_reset)
        self.sub_robot_mpc_x_plan = rospy.Subscriber('/husky/visualization/mpc_x_plan', Float64MultiArray,
                                                     self.receive_robot_mpc_x_plan)
        # self.sub_robot_info = rospy.Subscriber('/husky/visualization/robot_info', Float64MultiArray, self.receive_robot_info)
        # self.sub_obs_info = rospy.Subscriber('/husky/visualization/obs_info', Float64MultiArray, self.receive_obs_info)
        # self.sub_box_info = rospy.Subscriber('/husky/visualization/box_info', Float64MultiArray, self.receive_box_info)

    def receive_sim_reset(self, sim_reset_msg):
        # Retrieve the reset msg
        self.sim_reset = sim_reset_msg.data
        if self.sim_reset:
            self.robot_mpc_plan = None
            self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))
            self.sim_reset = False

    def receive_robot_mpc_x_plan(self, mpc_x_plan_msg):
        # save the msg to robot_mpc_x_plan
        self.robot_mpc_x_plan = np.array(mpc_x_plan_msg.data)  # column vector
        self.robot_mpc_x_plan = self.robot_mpc_x_plan.reshape((2, self.pr.N+1), order='C')  # matrix
        self.robot_mpc_plan = self.robot_mpc_x_plan

    def receive_robot_state(self, odom_robot):

        r = R.from_quat([odom_robot.pose.pose.orientation.x, 
                            odom_robot.pose.pose.orientation.y, 
                            odom_robot.pose.pose.orientation.z, 
                            odom_robot.pose.pose.orientation.w])

        ypr = r.as_euler('zyx', degrees=False)[0]

        vel = np.sqrt(odom_robot.twist.twist.linear.x**2 + odom_robot.twist.twist.linear.y**2)
        omega = odom_robot.twist.twist.angular.z
        self.robot_state = np.array([odom_robot.pose.pose.position.x, 
                                   odom_robot.pose.pose.position.y, ypr, vel, omega])

    def receive_object_state(self, obj_pos_msg):

        r = R.from_quat([obj_pos_msg.pose.pose.orientation.x, 
                            obj_pos_msg.pose.pose.orientation.y, 
                            obj_pos_msg.pose.pose.orientation.z, 
                            obj_pos_msg.pose.pose.orientation.w])

        ypr = r.as_euler('zyx', degrees=False)[0]

        vel_x = obj_pos_msg.twist.twist.linear.x
        vel_y = obj_pos_msg.twist.twist.linear.y
        omega = obj_pos_msg.twist.twist.angular.z
        self.object_state = np.array([obj_pos_msg.pose.pose.position.x, obj_pos_msg.pose.pose.position.y, ypr, vel_x, vel_y, omega]
)
    def prepare_visualization(self):
        # ================== prepare for visualization ====================
        # Prepare a figure for visualization 
        plt.ion()
        self.fig_main, self.ax_main = plt.subplots()
        self.fig_main.canvas.set_window_title('visualization in World frame')
        self.ax_main.grid(visible=True, ls='-.')
        self.ax_main.set_aspect('equal')
        self.ax_main.set_xlim(self.pr.ws_x)
        self.ax_main.set_ylim(self.pr.ws_y)
        self.ax_main.set_xlabel('x [m]')
        self.ax_main.set_ylabel('y [m]')

        # plot objects
        # robot current pos
        rob_corner_o = np.array([self.W_r, self.L_r]).T / -2.0   # position of the corners in the object frame
        rect_rob_pos_rec = mpatches.Rectangle([rob_corner_o[0], rob_corner_o[1]], self.W_r, self.L_r, 0.0, fc=(0, 1, 1, 0.1), ec=(0, 0, 1, 0.8))
        self.fig_robot_pos = self.ax_main.add_artist(rect_rob_pos_rec)

        # robot current orientation
        point_rd = np.dot(utils.rot_M(self.robot_state[2]), np.array([2.0, 0]).T) + np.array([self.robot_state[0], self.robot_state[1]])
        self.line_robot_direction = self.ax_main.plot([self.robot_state[0], point_rd[0]], [self.robot_state[1], point_rd[1]], linestyle = 'dotted', linewidth=2)

        # object current pos
        O_pos_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        object_pos_rec = mpatches.Rectangle([O_pos_corner_o[0], O_pos_corner_o[1]], self.W_o, self.L_o, 0.0, fc=(1, 1, 1, 0.1), ec=(1, 0, 1, 0.8))
        self.fig_object_pos = self.ax_main.add_artist(object_pos_rec)

        # robot history path 
        self.fig_robot_path = self.ax_main.plot(0.0, 0.0, c='g', ls='-', lw=2.0)

        # the predicted final state
        # W_posn_o_cir = mpatches.Circle(np.array([self.W_posn_o[0],self.W_posn_o[1]]), 0.1, fc=(1, 0.4, 1, 0.8), ec=(0, 0, 1, 0.8), alpha = 0.4)
        # self.fig_W_posn_o = self.ax_main.add_artist(W_posn_o_cir)

        # goal
        W_goal_cir = mpatches.Circle(np.array([self.obj_pos_goal[0],self.obj_pos_goal[1]]), 0.1, fc=(1, 0.1, 1, 0.8), ec=(0, 1, 1, 0.8), alpha = 1.)
        self.fig_W_goal_cir = self.ax_main.add_artist(W_goal_cir)

        plt.draw() 


    def render(self):
        # update robot hostory path
        self.rob_his_path = np.concatenate((self.rob_his_path, np.array(self.robot_state[0:3]).reshape((-1, 1))),
                                                 axis=1)
                                                 
        # Render the environment to the screen
        # Update robot current pos 
        rob_state = self.robot_state
        rob_corner_o = np.array([self.W_r, self.L_r]).T / -2.0   # position of the corners in the object frame
        rob_corner_w = rob_state[0:2].T + np.dot(utils.rot_M(rob_state[2]), rob_corner_o)      # position of the corner in the world frame
        self.fig_robot_pos.set_xy([rob_corner_w[0], rob_corner_w[1]])     # the left bottom coner position 
        self.fig_robot_pos.angle = np.rad2deg(rob_state[2])
        self.fig_robot_pos.set_width(self.W_r)
        self.fig_robot_pos.set_height(self.L_r)

        # update robot current orientation
        point_rd = np.dot(utils.rot_M(rob_state[2]), np.array([1.0, 0]).T) + np.array([rob_state[0], rob_state[1]])
        self.line_robot_direction[0].set_data([rob_state[0], point_rd[0]], [rob_state[1], point_rd[1]])

        # Update object current pos 
        O_pos_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corner in the object frame
        W_pos_corner_O = self.object_state[0:2] + np.dot(utils.rot_M(self.object_state[2]), O_pos_corner_o)      # position of the corner in the world frame
        self.fig_object_pos.set_xy([W_pos_corner_O[0], W_pos_corner_O[1]])     # the left bottom coner position 
        self.fig_object_pos.angle = np.rad2deg(self.object_state[2])
        self.fig_object_pos.set_width(self.W_o)
        self.fig_object_pos.set_height(self.L_o)

        # Update robot history path
        self.fig_robot_path[0].set_data(np.concatenate((self.rob_his_path[0, :], self.rob_his_path[1, :])).reshape((2, -1)))
        
        # Update the predicted final state
        # self.fig_W_posn_o.set_center([self.W_posn_o[0], self.W_posn_o[1]])

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()

    def init_path_candidates_both(self, N):
        self.mpc_planned_path = ['path_i']*N
        self.mpc_planned_path_object = ['path_i']*N
        for i in range(N):
            self.mpc_planned_path[i] = self.ax_main.plot(0.0, 0.0, c="red", ls='-', lw=2.0)
            self.mpc_planned_path_object[i] = self.ax_main.plot(0.0, 0.0, c="magenta", ls='-', lw=2.0)

    def show_paths_candidates_robot(self):
        for j in range(len(self.mpc_planned_path)):
            self.mpc_planned_path[j][0].set_data(self.robot_mpc_plan.reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()

def run_visualization_simple():
    print("Starting Visualization...")

    # Initialize a ros node
    rospy.init_node('husky_visualization_node', anonymous=False)

    # Specify the control node frequency
    hz = 20
    rate = rospy.Rate(hz)
    rospy.sleep(1.0)

    # Create a husky visualization object
    husky_vis = Husky_Visualization()
    husky_vis.rob_his_path = np.array(husky_vis.robot_state[0:3]).reshape((-1, 1))

    husky_vis.prepare_visualization()
    husky_vis.init_path_candidates_both(1)
    husky_vis.rob_his_path = np.array(husky_vis.robot_state[0:3]).reshape((-1, 1))

    # Start running
    n_loop = 0
    while not rospy.is_shutdown():
        husky_vis.render()
        husky_vis.show_paths_candidates_robot()
        
        n_loop += 1
        # print('vis {}'.format(n_loop))
        rate.sleep()

    plt.ioff()
    plt.show()


if __name__ == "__main__":
    run_visualization_simple()