import gym
from gym import spaces

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Arc, RegularPolygon
from scipy import integrate
from scipy.interpolate import interp1d
import math

from utils.integrator import my_RK4
from general_problem_setup import Problem, Index
import utils.utils as utils
from dynamics.rob_navi_dynamics import rob_navi_dynamics

class Rob_Navi(gym.Env):
    """Custom Environment that follows gym interface"""
    metadata = {'render.modes': ['human'], "render_fps": 30}

    def __init__(self):

        self.pr = Problem()
        self.index = Index()
        self.W_o = self.pr.rect_rob_size[0]
        self.L_o = self.pr.rect_rob_size[1]

        self.diata_t = 0.1

        # system setup
        self.r_r = self.pr.rob_size    # m (radius of the cylinder)

        # state initialization
        # state is for z
        self.state = np.random.uniform([-2.4, -2.4,        0.0, -self.pr.robot_maxVel*0.1, -self.pr.robot_maxOmega*0.01], \
                                       [0.0,  2.4,   np.pi/2.0,  self.pr.robot_maxVel*0.1,  self.pr.robot_maxOmega*0.01])

        # visualization initialization, range of the x-y axises
        self.ws_x = self.pr.ws_x
        self.ws_y = self.pr.ws_y

        # prepare for visualization
        self.prepare_visualization()

        self.done = False

    def step(self, action):
        # system state update
        # d_state = cylinder_pushing_dynamics(self.state, action, params=NAN)
        self.state = my_RK4(np.array(self.state),\
                            np.array(action),\
                            rob_navi_dynamics, self.diata_t, self.pr)

        reward = 0
        done = False 

        return np.array(self.state, dtype=np.float32), reward, done, {}

    def reset(self):
        # Reset the state of the environment to an initial state
        # state = [x_z, y_z, theta_z, v_z, d_theta_z]
        self.state = np.random.uniform([-2.4, -2.4,        0.0, -self.pr.robot_maxVel*0.1, -self.pr.robot_maxOmega*0.01], \
                                       [-1.2,  2.4,   np.pi/2.0,  self.pr.robot_maxVel*0.1,  self.pr.robot_maxOmega*0.01])

        # initialize object history path
        self.rob_his_path = np.array(self.state[0:3]).reshape((-1, 1))

        self.done = False

        return self.state

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        # Update robot current pos 
        self.rob_state = self.state
        rob_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        rob_corner_w = self.rob_state[0:2].T + np.dot(utils.rot_M(self.rob_state[2]), rob_corner_o)      # position of the corner in the world frame
        self.fig_robot_pos.set_xy([rob_corner_w[0], rob_corner_w[1]])     # the left bottom coner position 
        self.fig_robot_pos.angle = np.rad2deg(self.rob_state[2])
        self.fig_robot_pos.set_width(self.W_o)
        self.fig_robot_pos.set_height(self.L_o)

        # update robot current orientation
        point_rd = np.dot(utils.rot_M(self.rob_state[2]), np.array([1.0, 0]).T) + np.array([self.rob_state[0], self.rob_state[1]])
        self.line_robot_direction[0].set_data([self.rob_state[0], point_rd[0]], [self.rob_state[1], point_rd[1]])

        # Update object history path
        self.fig_object_path[0].set_data(np.concatenate((self.rob_his_path[0, :], self.rob_his_path[1, :])).reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()
        
    def prepare_visualization(self):
        # ================== prepare for visualization ====================
        # Prepare a figure for visualization 
        plt.ion()
        self.fig_main, self.ax_main = plt.subplots()
        self.ax_main.grid(visible=True, ls='-.')
        self.ax_main.set_aspect('equal')
        self.ax_main.set_xlim(self.ws_x)
        self.ax_main.set_ylim(self.ws_y)
        self.ax_main.set_xlabel('x [m]')
        self.ax_main.set_ylabel('y [m]')

        # plot objects
        # robot current pos
        rob_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        rect_rob_pos_rec = mpatches.Rectangle([rob_corner_o[0], rob_corner_o[1]], self.W_o, self.L_o, 0.0, fc=(0, 1, 1, 0.1), ec=(0, 0, 1, 0.8))
        self.fig_robot_pos = self.ax_main.add_artist(rect_rob_pos_rec)

        # robot current orientation
        point_rd = np.dot(utils.rot_M(self.state[2]), np.array([2.0, 0]).T) + np.array([self.state[0], self.state[1]])
        self.line_robot_direction = self.ax_main.plot([self.state[0], point_rd[0]], [self.state[1], point_rd[1]], linestyle = 'dotted', linewidth=2)

        # object history path 
        self.fig_object_path = self.ax_main.plot(0.0, 0.0, c='g', ls='-', lw=2.0)
        plt.draw() 

    def close(self):
        plt.close(self.fig_main)

    def init_path_candidates_both(self, N):
        self.mpc_planned_path = ['path_i']*N
        self.mpc_planned_path_object = ['path_i']*N
        for i in range(N):
            self.mpc_planned_path[i] = self.ax_main.plot(0.0, 0.0, c="red", ls='-', lw=2.0)
            self.mpc_planned_path_object[i] = self.ax_main.plot(0.0, 0.0, c="magenta", ls='-', lw=2.0)

    def show_paths_candidates_robot(self, mpc_path):
        all_x_coord = []
        all_y_coord = []
        all_x_coord_object = []
        all_y_coord_object = []

        for i in range(0, mpc_path.shape[1]):
            x_coord = mpc_path[:, i, 0]  # a list of x in one episode
            y_coord = mpc_path[:, i, 1]  # a list of y in one episode
            all_x_coord.append(x_coord)
            all_y_coord.append(y_coord)

        # # first remove all paths
        # for paths in self.mpc_planned_path:
        #     paths.remove()
        # now draw new planned paths
        for j in range(len(self.mpc_planned_path)):
            # self.mpc_planned_path.append(self.ax_main.plot(all_x_coord[j], all_y_coord[j], c="red", ls='-', lw=2.0))
            self.mpc_planned_path[j][0].set_data(np.concatenate((all_x_coord[j], all_y_coord[j])).reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()

    def show_paths_candidates_both(self, mpc_path):
        all_x_coord = []
        all_y_coord = []
        all_x_coord_object = []
        all_y_coord_object = []

        for i in range(0, mpc_path.shape[1]):
            x_coord = mpc_path[:, i, 0]  # a list of x in one episode
            y_coord = mpc_path[:, i, 1]  # a list of y in one episode
            all_x_coord.append(x_coord)
            all_y_coord.append(y_coord)

            x_coord_object = mpc_path[:, i, 5]  # a list of x in one episode
            y_coord_object = mpc_path[:, i, 6]  # a list of y in one episode
            all_x_coord_object.append(x_coord_object)
            all_y_coord_object.append(y_coord_object)

        # # first remove all paths
        # for paths in self.mpc_planned_path:
        #     paths.remove()
        # now draw new planned paths
        for j in range(len(self.mpc_planned_path)):
            # self.mpc_planned_path.append(self.ax_main.plot(all_x_coord[j], all_y_coord[j], c="red", ls='-', lw=2.0))
            self.mpc_planned_path[j][0].set_data(np.concatenate((all_x_coord[j], all_y_coord[j])).reshape((2, -1)))
            self.mpc_planned_path_object[j][0].set_data(np.concatenate((all_x_coord_object[j], all_y_coord_object[j])).reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()

    def save_trajectory(self):
        # action [acc_r, acc_omega]
        # robot state: self.rob_state [x, y, theta, vr, omega]
        # object state: self.obj_state [x, y, theta, vx, vy, omega]

        full_state = np.append(np.append(np.append(self.prev_rob_state, self.prev_obj_state), self.action), self.time)
        full_state = full_state.reshape(1, -1)
        self.full_trajectory = np.concatenate((self.full_trajectory, full_state), axis=0)

    def save_trajectory_data(self, i_episode, save_directory):
        # save_directory = '/home/susan/Documents/_isaac/sim_yu/trajectories/trajectories_collect/push_trajectory'
        np.save(save_directory + str(i_episode) + '.npy', self.full_trajectory)
