import gym
from gym import spaces, GoalEnv

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Arc, RegularPolygon
from scipy import integrate
from scipy.interpolate import interp1d
import math
from collections import OrderedDict

from utils.integrator import my_RK4
from general_problem_setup import Problem, Index
import utils.utils as utils
import utils.Rectangle as Rectangle

from dynamics.rob_navi_dynamics import rob_navi_dynamics
from controllers.mpc.stable_push_mpc_controller import Stable_push_mpc_controller


class Rob_Navi(GoalEnv):
    """Custom Environment that follows gym interface"""
    metadata = {'render.modes': ['human'], "render_fps": 30}

    def __init__(self):

        self.pr = Problem()
        self.index = Index()

        self.diata_t = 0.1
        self.t = 0

        # system setup
        self.W_o = self.pr.rect_rob_size[0]
        self.L_o = self.pr.rect_rob_size[1]

        # observation and action space
        self.high = np.array([2.4,  2.4,  np.pi/1.0,  self.pr.robot_maxVel*0.1,  self.pr.robot_maxOmega*0.01])
        self.low = np.array([-2.4, -2.4, -np.pi/1.0, -self.pr.robot_maxVel*0.1, -self.pr.robot_maxOmega*0.01])
        # dictionary observations refer to https://github.com/DLR-RM/stable-baselines3/pull/243
        # contact mode (with regard to the box)
        # contact_state = 0, contact at the left of the box, the left side is the one y axis points to
        # contact_state = 1, contact at the back side of the box
        # contact_state = 2, contact at the right side of the box
        # contact_state = 3, contact at the front side of the box
        self.observation_space  = spaces.Dict({
            'robot_state': spaces.Box(low=self.low, high=self.high, dtype=np.float32),
            'contact_state': spaces.Discrete(4), 
        })
        # self.observation_space = spaces.Dict(observation_space)
        # action = 0,  rotate 90 degrees counterclockwise
        # action = 1,  do not switch contact
        # action = 2,  rotate 90 degrees clockwise
        self.action_space = spaces.Discrete(3)  

        # state initialization
        self.robot_state = np.random.uniform(self.low, self.high)
        # self.obj_pos_goal = np.random.uniform(self.low[0:2], self.high[0:2])
        self.contact_state = 1
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', self.contact_state)
        ])

        # visualization initialization, range of the x-y axises
        self.ws_x = self.pr.ws_x
        self.ws_y = self.pr.ws_y

        # Initializing robot navi mpc casadi
        self.mpc_controller = Stable_push_mpc_controller()
        recompile = True
        solver_build_name = 'Stable_push_mpc_controller_casadi'    
        self.mpc_controller.generate_solver(recompile, solver_build_name)
        # Initializing the mpc related values
        mpc_x_plan = np.tile(np.array(self.robot_state).reshape((-1, 1)), (1, self.pr.N))
        mpc_u_plan = np.zeros((self.pr.nu, self.pr.N))
        self.mpc_feasible = False 

        # initialize object history path
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))

        # task parameters
        self.obj_pos_goal = self.pr.obj_pos_goal
        self.obj_theta_goal = self.pr.obj_theta_goal
        self.v_omega_k_max = [0.3, 0.2, 0.2, 0.3]     # [left, back, right, front]
        self.v_omega_k_min = [-0.3, -0.2, -0.2, -0.3]     # [left, back, right, front]

        # obstacle setup
        obs_pos_w = [-1, 1]
        obs_angle = np.pi/3
        obs_size = [1.5, 0.4]
        self.obs_rect = Rectangle.Rectangle(obs_pos_w, obs_angle, obs_size)
        self.robot_rect = Rectangle.Rectangle(self.robot_state[0:2], self.robot_state[2], [self.W_o, self.L_o])

        # prepare for visualization
        self.prepare_visualization()

        self.done = False
        self.render()

    def step(self, action):
        self.t = self.t + self.diata_t

        # get current state from observation
        self.robot_state = self.observation['robot_state']
        self.contact_state = self.observation['contact_state']

        # contact state update
        self.contact_state = self.contact_state + (action-1)
        if self.contact_state < 0:
            self.contact_state = 3
        elif self.contact_state > 3:
            self.contact_state = 0
        
        # robot state update (rotation when switching the contact)
        self.robot_state[2] = self.robot_state[2] + (action-1)*np.pi/2
        angle = self.robot_state[2]%(2*np.pi)
        self.robot_state[2] = angle if angle < np.pi else angle - 2*np.pi

        # low lever control from the mpc controller
        self.mpc_controller.set_system_state_current(self.robot_state)
        self.mpc_controller.set_obj_pos_theta_goal(self.obj_pos_goal, self.obj_theta_goal[0])
        self.mpc_controller.run_mpc(self.v_omega_k_max[self.contact_state], self.v_omega_k_min[self.contact_state])
        self.mpc_feasible = self.mpc_controller.mpc_feasible

        # robot state update
        # d_state = cylinder_pushing_dynamics(self.robot_state, action, params=NAN)
        self.last_robot_state = self.robot_state
        self.robot_state = my_RK4(np.array(self.robot_state),\
                            np.array(self.mpc_controller.robot_acc_omega_acc_current),\
                            rob_navi_dynamics, self.diata_t, self.pr)
        self.robot_rect.update_rectangle(self.robot_state[0:2], self.robot_state[2])

        # update visualization
        self.rob_his_path = np.concatenate((self.rob_his_path, np.array(self.robot_state[0:3]).reshape((-1, 1))),
                                                 axis=1)

        self.show_paths_candidates_robot(self.mpc_controller.system_state_plan)
        self.render()

        # update observation
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', self.contact_state)
        ])

        # to maximize the reward
        disGoal = np.linalg.norm(self.robot_state[0:2] - self.obj_pos_goal)
        # rewards
        closeToTheGoalReward = - (np.linalg.norm(self.robot_state[0:2] - self.obj_pos_goal)\
                                    - np.linalg.norm(self.last_robot_state[0:2] - self.obj_pos_goal))
        timeCost = -1
        # traveledDisCost = - np.linalg.norm(self.last_robot_state[0:2] - self.robot_state[0:2])
        contactSwitchingCost = - abs((action-1))
        # weights of the rewards
        w_goal = 50
        w_time = 1
        w_contactSwitching = 1

        reward = w_goal*closeToTheGoalReward + w_time*timeCost + w_contactSwitching*contactSwitchingCost

        # check occupancy
        self.collision_state = False
        self.collision_state = Rectangle.collideRectPolygon(self.obs_rect, self.robot_rect)

        success = False
        if disGoal < 0.5:
            reward = 200
            self.done = True
            success = True
        elif self.t > 25:
            reward = - 50
            self.done = True
        elif self.collision_state == True:
            reward = - 200
            self.done = True
        else:
            self.done = False

        return self.observation, reward, self.done, {'success':success}

    def reset(self):
        self.t = 0

        # Reset the state of the environment to an initial state
        # state = [x_z, y_z, theta_z, v_z, d_theta_z]
        self.robot_state = np.random.uniform(self.low, self.high)
        self.contact_state = 1
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', self.contact_state)
        ])

        # initialize object history path
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))

        self.done = False
        self.render()

        return self.observation

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        # Update robot current pos 
        rob_state = self.robot_state

        self.rob_corners_o = np.array([[self.W_o/ -2.0, self.L_o/ -2.0], 
                                       [self.W_o/  2.0, self.L_o/ -2.0], 
                                       [self.W_o/  2.0, self.L_o/  2.0], 
                                       [self.W_o/ -2.0, self.L_o/  2.0]]).T    # position of the corners in the object frame
        self.rob_corners_wr = np.dot(utils.rot_M(rob_state[2]), self.rob_corners_o)      # position of the corner in the world frame
        self.rob_corners_w = self.rob_corners_wr + np.array([[rob_state[0]],[rob_state[1]]])

        rob_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        rob_corner_w = rob_state[0:2].T + np.dot(utils.rot_M(rob_state[2]), rob_corner_o)      # position of the corner in the world frame
        self.fig_robot_pos.set_xy([rob_corner_w[0], rob_corner_w[1]])     # the left bottom coner position 
        self.fig_robot_pos.angle = np.rad2deg(rob_state[2])
        self.fig_robot_pos.set_width(self.W_o)
        self.fig_robot_pos.set_height(self.L_o)

        # update robot current orientation
        point_rd = np.dot(utils.rot_M(rob_state[2]), np.array([1.0, 0]).T) + np.array([rob_state[0], rob_state[1]])
        self.line_robot_direction[0].set_data([rob_state[0], point_rd[0]], [rob_state[1], point_rd[1]])

        # Update object history path
        self.fig_object_path[0].set_data(np.concatenate((self.rob_his_path[0, :], self.rob_his_path[1, :])).reshape((2, -1)))

        # Update object position goal
        self.fig_goal_pos.set_center([self.obj_pos_goal[0], self.obj_pos_goal[1]])

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()
        
    def prepare_visualization(self):
        # ================== prepare for visualization ====================
        # Prepare a figure for visualization 
        plt.ion()
        self.fig_main, self.ax_main = plt.subplots()
        self.ax_main.grid(visible=True, ls='-.')
        self.ax_main.set_aspect('equal')
        self.ax_main.set_xlim(self.ws_x)
        self.ax_main.set_ylim(self.ws_y)
        self.ax_main.set_xlabel('x [m]')
        self.ax_main.set_ylabel('y [m]')

        # plot objects
        # robot current pos
        rob_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        rect_rob_pos_rec = mpatches.Rectangle([rob_corner_o[0], rob_corner_o[1]], self.W_o, self.L_o, 0.0, fc=(0, 1, 1, 0.1), ec=(0, 0, 1, 0.8))
        self.fig_robot_pos = self.ax_main.add_artist(rect_rob_pos_rec)

        # robot current orientation
        point_rd = np.dot(utils.rot_M(self.robot_state[2]), np.array([2.0, 0]).T) + np.array([self.robot_state[0], self.robot_state[1]])
        self.line_robot_direction = self.ax_main.plot([self.robot_state[0], point_rd[0]], [self.robot_state[1], point_rd[1]], linestyle = 'dotted', linewidth=2)

        # object history path 
        self.fig_object_path = self.ax_main.plot(0.0, 0.0, c='g', ls='-', lw=2.0)

        # obstacle
        obs_corner_o = np.array([self.obs_rect.W_o, self.obs_rect.L_o]).T / -2.0   # position of the corners in the object frame
        obs_corner_w = self.obs_rect.center + np.dot(utils.rot_M(self.obs_rect.angle), obs_corner_o)      # position of the corner in the world frame
        obstacle_rec = mpatches.Rectangle(obs_corner_w, self.obs_rect.W_o, self.obs_rect.L_o, np.rad2deg(self.obs_rect.angle), fc=(1, 0, 1, 0.1), ec=(1, 0, 1, 0.8))
        self.fig_obstacle = self.ax_main.add_artist(obstacle_rec)

        # goal point
        goal_cir = mpatches.Circle(np.array([self.obj_pos_goal[0], self.obj_pos_goal[1]]), 0.2, fc=(1, 0.1, 1, 0.5), ec=(1, 0, 0, 0.8), alpha = 0.4)
        self.fig_goal_pos = self.ax_main.add_artist(goal_cir)

        plt.draw() 

    def close(self):
        plt.close(self.fig_main)

    def init_path_candidates_both(self, N):
        self.mpc_planned_path = ['path_i']*N
        self.mpc_planned_path_object = ['path_i']*N
        for i in range(N):
            self.mpc_planned_path[i] = self.ax_main.plot(0.0, 0.0, c="red", ls='-', lw=2.0)
            self.mpc_planned_path_object[i] = self.ax_main.plot(0.0, 0.0, c="magenta", ls='-', lw=2.0)

    def show_paths_candidates_robot(self, mpc_path):
        all_x_coord = []
        all_y_coord = []
        all_x_coord_object = []
        all_y_coord_object = []

        for i in range(0, mpc_path.shape[1]):
            x_coord = mpc_path[:, i, 0]  # a list of x in one episode
            y_coord = mpc_path[:, i, 1]  # a list of y in one episode
            all_x_coord.append(x_coord)
            all_y_coord.append(y_coord)

        for j in range(len(self.mpc_planned_path)):
            self.mpc_planned_path[j][0].set_data(np.concatenate((all_x_coord[j], all_y_coord[j])).reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()


    def save_trajectory(self):
        # action [acc_r, acc_omega]
        # robot state: rob_state [x, y, theta, vr, omega]
        # object state: self.obj_state [x, y, theta, vx, vy, omega]

        full_state = np.append(np.append(np.append(self.prev_rob_state, self.prev_obj_state), self.action), self.time)
        full_state = full_state.reshape(1, -1)
        self.full_trajectory = np.concatenate((self.full_trajectory, full_state), axis=0)

    def save_trajectory_data(self, i_episode, save_directory):
        # save_directory = '/home/susan/Documents/_isaac/sim_yu/trajectories/trajectories_collect/push_trajectory'
        np.save(save_directory + str(i_episode) + '.npy', self.full_trajectory)
