import gym
# from gym import spaces, GoalEnv

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Arc, RegularPolygon
from scipy import integrate
from scipy.interpolate import interp1d
import math
from collections import OrderedDict

from utils.integrator import my_RK4
from general_problem_setup import Problem, Index
import utils.utils as utils
from dynamics.rob_navi_dynamics import rob_navi_dynamics
from controllers.mpc.stable_push_mpc_controller import Stable_push_mpc_controller


class Stable_push_with_contact_switching_vis():
    """Custom Environment that follows gym interface"""
    metadata = {'render.modes': ['human'], "render_fps": 30}

    def __init__(self):

        self.pr = Problem()
        self.index = Index()

        self.diata_t = 0.1
        self.t = 0

        # system setup
        self.W_r = self.pr.rect_rob_size[0]
        self.L_r = self.pr.rect_rob_size[1]
        self.W_o = self.pr.obj_size[0]
        self.L_o = self.pr.obj_size[1]

        # observation and action space
        self.high = np.array([2.4,  2.4,  np.pi/1.0,  self.pr.robot_maxVel*0.1,  self.pr.robot_maxOmega*0.01])
        self.low = np.array([-2.4, -2.4, -np.pi/1.0, -self.pr.robot_maxVel*0.1, -self.pr.robot_maxOmega*0.01])
        # dictionary observations refer to https://github.com/DLR-RM/stable-baselines3/pull/243
        # contact mode (with regard to the box)
        # contact_state = 0, contact at the left of the box, the left side is the one y axis points to
        # contact_state = 1, contact at the back side of the box
        # contact_state = 2, contact at the right side of the box
        # contact_state = 3, contact at the front side of the box
        # self.observation_space  = spaces.Dict({
        #     'robot_state': spaces.Box(low=self.low, high=self.high, dtype=np.float32),
        #     'contact_state': spaces.Discrete(4), 
        # })
        # self.observation_space = spaces.Dict(observation_space)
        # action = 0,  rotate 90 degrees counterclockwise
        # action = 1,  do not switch contact
        # action = 2,  rotate 90 degrees clockwise
        # action = 3,  rotate 180 degrees clockwise
        # self.action_space = spaces.Discrete(4)  

        # state initialization
        self.robot_state = np.random.uniform(self.low, self.high)
        contact_state = 1
        # contact_state = self.observation_space['contact_state'].sample()
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', contact_state)
        ])

        # object coordinate with reference to the robot frame, according to the contact state
        self.list_R_pose_o = np.array( [[(self.L_o + self.W_r)/2, 0, np.pi/2],    # 'contact_state_0'
                                        [(self.W_o + self.W_r)/2, 0, 0],          # 'contact_state_1'
                                        [(self.L_o + self.W_r)/2, 0, -np.pi/2],   # 'contact_state_2'
                                        [(self.W_o + self.W_r)/2, 0, np.pi]    ]) # 'contact_state_3'
        self.dist = 0.3
        self.list_R_pose_o_withDist = np.array( [[(self.L_o + self.W_r)/2 + self.dist, 0, np.pi/2],    # 'contact_state_0'
                                        [(self.W_o + self.W_r)/2 + self.dist, 0, 0],          # 'contact_state_1'
                                        [(self.L_o + self.W_r)/2 + self.dist, 0, -np.pi/2],   # 'contact_state_2'
                                        [(self.W_o + self.W_r)/2 + self.dist, 0, np.pi]    ]) # 'contact_state_3'
        # object state initialization
        R_pos_o = self.list_R_pose_o[self.observation['contact_state']]
        W_pos_o = np.dot(utils.rot_M(self.robot_state[2]), np.array(R_pos_o[0:2]).T) + np.array([self.robot_state[0], self.robot_state[1]])
        W_angle_o = self.robot_state[2] + R_pos_o[2]
        self.object_state = np.append(W_pos_o,W_angle_o)

        # visualization initialization, range of the x-y axises
        self.ws_x = self.pr.ws_x
        self.ws_y = self.pr.ws_y

        # prepare for visualization
        self.W_posn_o = np.array([0., 0.])
        self.last_W_posn_o = np.array([0., 0.])

        # Initializing robot navi mpc casadi
        self.mpc_controller = Stable_push_mpc_controller()
        recompile = True
        solver_build_name = 'Stable_push_mpc_controller_casadi'    
        self.mpc_controller.generate_solver(recompile, solver_build_name)
        # Initializing the mpc related values
        mpc_x_plan = np.tile(np.array(self.robot_state).reshape((-1, 1)), (1, self.pr.N))
        mpc_u_plan = np.zeros((self.pr.nu, self.pr.N))
        self.mpc_feasible = False 

        # initialize object history path
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))

        # task parameters
        self.obj_pos_goal = self.pr.obj_pos_goal
        self.obj_theta_goal = self.pr.obj_theta_goal
        self.v_omega_k_max = [0.3, 0.2, 0.2, 0.3]           # [left, back, right, front]
        self.v_omega_k_min = [-0.3, -0.2, -0.2, -0.3]       # [left, back, right, front]
        self.cost_map = np.zeros([60,60])

        self.done = False
        self.prepare_visualization()
        self.render()

    # as a visualization tool


    def step(self, action):
        self.t = self.t + self.diata_t

        # get current state from observation
        self.robot_state = self.observation['robot_state']
        contact_state = self.observation['contact_state']

        # contact state update
        contact_state = contact_state - (action-1)
        if contact_state == -1:
            contact_state = 3
        elif contact_state == -2:
            contact_state = 2
        elif contact_state == 4:
            contact_state = 0
        elif contact_state == 5:
            contact_state = 1
        
        # robot state update 
        self.last_robot_state = self.robot_state
        self.last_W_posn_o = self.W_posn_o
        R_pos_o = self.list_R_pose_o[contact_state]
        R_pos_o_withDist = self.list_R_pose_o_withDist[contact_state]

        if action == 1:
            # low lever control from the mpc controller
            self.mpc_controller.set_system_state_current(self.robot_state)
            self.mpc_controller.set_obj_pos_theta_goal(self.obj_pos_goal, self.obj_theta_goal[0])
            self.mpc_controller.run_mpc(self.v_omega_k_max[contact_state], self.v_omega_k_min[contact_state], R_pos_o)
            self.mpc_feasible = self.mpc_controller.mpc_feasible

            # d_state = cylinder_pushing_dynamics(self.robot_state, action, params=NAN)
            self.robot_state = my_RK4(np.array(self.robot_state),\
                                np.array(self.mpc_controller.robot_acc_omega_acc_current),\
                                rob_navi_dynamics, self.diata_t, self.pr)

            # object state update
            W_pos_o = np.dot(utils.rot_M(self.robot_state[2]), np.array(R_pos_o[0:2]).T) + \
                    np.array([self.robot_state[0], self.robot_state[1]])
            self.object_state[0:2] = W_pos_o    
            W_angle_o = self.robot_state[2] + R_pos_o[2]  
            self.object_state[2] = W_angle_o   

            # the predicted final state
            robot_state_n = self.mpc_controller.system_state_plan[-1, 0, 0:3]    # mpc planned xn_r
            # yn_r = self.mpc_controller.system_state_plan[:, -1, 1]    # mpc planned xn_r
            self.W_posn_o = np.dot(utils.rot_M(robot_state_n[2]), np.array(R_pos_o[0:2]).T) + \
                    np.array([robot_state_n[0], robot_state_n[1]])          

        else:
            self.robot_state[3] = 0.
            self.robot_state[4] = 0.
            
            # rotation when switching the contact
            # next contact point
            robot_next_contact_state = np.array([0.0, 0.0, 0.0])
            robot_next_contact_state[2] = self.robot_state[2] - (action-1)*np.pi/2
            angle = robot_next_contact_state[2]%(2*np.pi)
            robot_next_contact_state[2] = angle if angle < np.pi else angle - 2*np.pi
            robot_next_contact_state[0:2] = self.object_state[0:2] - np.dot(utils.rot_M(robot_next_contact_state[2]), np.array(R_pos_o[0:2]))

            # first reverse 0.5m
            robot_state_pre = self.robot_state
            while utils.distPoints(robot_state_pre[0], robot_state_pre[1], self.robot_state[0], self.robot_state[1]) < self.dist:
                # reverse at a speed of -0.1 m/s
                self.robot_state[3] = -0.1
                d_x_r = np.cos(self.robot_state[2])*self.robot_state[3]*self.diata_t
                d_y_r = np.sin(self.robot_state[2])*self.robot_state[3]*self.diata_t
                self.robot_state[0] = self.robot_state[0] + d_x_r
                self.robot_state[1] = self.robot_state[1] + d_y_r
                self.render()
            
            # run feedback control, go to the prepare_to_push spot
            # the prepare_to_push spot is:
            rob_sub_goal = np.array([0.0, 0.0, 0.0])
            rob_sub_goal[2] = rob_sub_goal[2] - (action-1)*np.pi/2
            angle = rob_sub_goal[2]%(2*np.pi)
            rob_sub_goal[2] = angle if angle < np.pi else angle - 2*np.pi
            rob_sub_goal[0:2] = self.object_state[0:2] - np.dot(utils.rot_M(rob_sub_goal[2]), np.array(R_pos_o_withDist[0:2]))
            # turn 90 degrees first
            while abs(self.robot_state[2] - robot_state_pre) > 85/180*np.pi:
                if action == 0:
                    self.robot_state[4] = -0.1
                elif action == 2:
                    self.robot_state[4] = 0.1
                self.robot_state[2] = self.robot_state[2] + self.robot_state[4]*self.diata_t
                self.render()

            # start feedback control
            while utils.distPoints(rob_sub_goal[0], rob_sub_goal[1], self.robot_state[0], self.robot_state[1]) < 0.1 and abs(self.robot_state[2] - rob_sub_goal) < 0.1:
                robot_control_current = self.run_feedback_obs_free_navi(self.robot_state, rob_sub_goal)
                self.robot_state[3] = robot_control_current[0]
                self.robot_state[4] = robot_control_current[1]
                d_x_r = np.cos(self.robot_state[2])*self.robot_state[3]*self.diata_t
                d_y_r = np.sin(self.robot_state[2])*self.robot_state[3]*self.diata_t
                self.robot_state[0] = self.robot_state[0] + d_x_r
                self.robot_state[1] = self.robot_state[1] + d_y_r
                self.robot_state[2] = self.robot_state[2] + self.robot_state[4]*self.diata_t
                self.render()

            # move forward to get in touch with the object
            while utils.distPoints(robot_next_contact_state[0], robot_next_contact_state[1], self.robot_state[0], self.robot_state[1]) < 0.05:
                # reverse at a speed of -0.1 m/s
                self.robot_state[3] = 0.05
                d_x_r = np.cos(self.robot_state[2])*self.robot_state[3]*self.diata_t
                d_y_r = np.sin(self.robot_state[2])*self.robot_state[3]*self.diata_t
                self.robot_state[0] = self.robot_state[0] + d_x_r
                self.robot_state[1] = self.robot_state[1] + d_y_r
                self.render()

            # the predicted final state
            self.W_posn_o = self.object_state[0:2]

        # update visualization
        self.rob_his_path = np.concatenate((self.rob_his_path, np.array(self.robot_state[0:3]).reshape((-1, 1))),
                                                 axis=1)

        self.show_paths_candidates_robot(self.mpc_controller.system_state_plan)
        self.render()

        # update observation
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', contact_state)
        ])

        # to maximize the reward
        disGoal = np.linalg.norm(self.object_state[0:2] - self.obj_pos_goal)
        # rewards
        closeToTheGoalReward = - (np.linalg.norm(self.robot_state[0:2] - self.obj_pos_goal)\
                                    - np.linalg.norm(self.last_robot_state[0:2] - self.obj_pos_goal))
        timeCost = -1
        # traveledDisCost = - np.linalg.norm(self.last_robot_state[0:2] - self.robot_state[0:2])
        if action != 1:
            action_cost = 1
        else:
            action_cost = 0
        contactSwitchingCost = - action_cost
        # mpc promising reward, xn_o - goal
        mpcPredictedReward = - (np.linalg.norm(self.W_posn_o[0:2] - self.obj_pos_goal)\
                                - np.linalg.norm(self.last_W_posn_o[0:2] - self.obj_pos_goal))

        # weights of the rewards
        w_goal = 80
        w_time = 2
        w_contactSwitching = 12#12
        w_mpcPred = 40

        reward = w_goal*closeToTheGoalReward + \
                 w_time*timeCost + \
                 w_contactSwitching*contactSwitchingCost + \
                 w_mpcPred*mpcPredictedReward

        # print("closeToTheGoalReward: ", w_goal*closeToTheGoalReward)
        # print("mpcPredictedReward: ", w_mpcPred*mpcPredictedReward)
        
        success = False
        if disGoal < 0.4:
            reward = 200
            self.done = True
            success = True
        elif self.t > 25:
            reward = - 100
            self.done = True
        elif self.robot_state[0] > self.pr.ws_x[1] or self.robot_state[0] < self.pr.ws_x[0] \
            or self.robot_state[1] > self.pr.ws_y[1] or self.robot_state[1] < self.pr.ws_y[0]:   # out of the arena
            reward = - 50
            self.done = True
        else:
            self.done = False

        return self.observation, reward, self.done, {'success':success}

    def reset(self):
        self.t = 0

        # Reset the state of the environment to an initial state
        # state = [x_z, y_z, theta_z, v_z, d_theta_z]
        self.robot_state = np.random.uniform(self.low, self.high)
        contact_state = self.observation_space['contact_state'].sample()
        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', contact_state)
        ])

        # object state initialization
        R_pos_o = self.list_R_pose_o[self.observation['contact_state']]
        W_pos_o = np.dot(utils.rot_M(self.robot_state[2]), np.array(R_pos_o[0:2]).T) + np.array([self.robot_state[0], self.robot_state[1]])
        W_angle_o = self.robot_state[2] + R_pos_o[2]
        self.object_state = np.append(W_pos_o, W_angle_o)

        # initialize object history path
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))
        # initialize the predicted final state
        self.W_posn_o = np.array([0., 0.])
        self.last_W_posn_o = np.array([0., 0.])

        self.done = False
        self.render()

        return self.observation

    def set_state(self, cur_rob_state, cur_obj_state, contact_state):

        self.object_state = cur_obj_state[0:3]
        self.robot_state = cur_rob_state

        self.observation = OrderedDict([
            ('robot_state', self.robot_state),
            ('contact_state', contact_state)
        ])

        # initialize object history path
        self.rob_his_path = np.array(self.robot_state[0:3]).reshape((-1, 1))

        self.done = False

    def render(self, mode='human', close=False):
        # Render the environment to the screen
        # Update robot current pos 
        rob_state = self.robot_state
        rob_corner_o = np.array([self.W_r, self.L_r]).T / -2.0   # position of the corners in the object frame
        rob_corner_w = rob_state[0:2].T + np.dot(utils.rot_M(rob_state[2]), rob_corner_o)      # position of the corner in the world frame
        self.fig_robot_pos.set_xy([rob_corner_w[0], rob_corner_w[1]])     # the left bottom coner position 
        self.fig_robot_pos.angle = np.rad2deg(rob_state[2])
        self.fig_robot_pos.set_width(self.W_r)
        self.fig_robot_pos.set_height(self.L_r)

        # update robot current orientation
        point_rd = np.dot(utils.rot_M(rob_state[2]), np.array([1.0, 0]).T) + np.array([rob_state[0], rob_state[1]])
        self.line_robot_direction[0].set_data([rob_state[0], point_rd[0]], [rob_state[1], point_rd[1]])

        # Update object current pos 
        O_pos_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corner in the object frame
        W_pos_corner_O = self.object_state[0:2] + np.dot(utils.rot_M(self.object_state[2]), O_pos_corner_o)      # position of the corner in the world frame
        self.fig_object_pos.set_xy([W_pos_corner_O[0], W_pos_corner_O[1]])     # the left bottom coner position 
        self.fig_object_pos.angle = np.rad2deg(self.object_state[2])
        self.fig_object_pos.set_width(self.W_o)
        self.fig_object_pos.set_height(self.L_o)

        # Update robot history path
        self.fig_robot_path[0].set_data(np.concatenate((self.rob_his_path[0, :], self.rob_his_path[1, :])).reshape((2, -1)))
        
        # Update the predicted final state
        # self.fig_W_posn_o.set_center([self.W_posn_o[0], self.W_posn_o[1]])

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()

    def init_goal_pos(self, goal_pos):
        goal_cir = mpatches.Circle(np.array([goal_pos[0],goal_pos[1]]), 0.2, fc=(1, 0.1, 1, 0.5), ec=(1, 0, 0, 0.8), alpha = 0.4)
        self.fig_goal_pos = self.ax_main.add_artist(goal_cir)

    def set_goal_pos(self, goal_pos):
        self.fig_goal_pos.set_center([goal_pos[0], goal_pos[1]])
        
    def prepare_visualization(self):
        # ================== prepare for visualization ====================
        # Prepare a figure for visualization 
        plt.ion()
        self.fig_main, self.ax_main = plt.subplots()
        self.fig_main.canvas.set_window_title('visualization in World frame')
        self.ax_main.grid(visible=True, ls='-.')
        self.ax_main.set_aspect('equal')
        self.ax_main.set_xlim(self.ws_x)
        self.ax_main.set_ylim(self.ws_y)
        self.ax_main.set_xlabel('x [m]')
        self.ax_main.set_ylabel('y [m]')

        # plot objects
        # robot current pos
        rob_corner_o = np.array([self.W_r, self.L_r]).T / -2.0   # position of the corners in the object frame
        rect_rob_pos_rec = mpatches.Rectangle([rob_corner_o[0], rob_corner_o[1]], self.W_r, self.L_r, 0.0, fc=(0, 1, 1, 0.1), ec=(0, 0, 1, 0.8))
        self.fig_robot_pos = self.ax_main.add_artist(rect_rob_pos_rec)

        # robot current orientation
        point_rd = np.dot(utils.rot_M(self.robot_state[2]), np.array([2.0, 0]).T) + np.array([self.robot_state[0], self.robot_state[1]])
        self.line_robot_direction = self.ax_main.plot([self.robot_state[0], point_rd[0]], [self.robot_state[1], point_rd[1]], linestyle = 'dotted', linewidth=2)

        # object current pos
        O_pos_corner_o = np.array([self.W_o, self.L_o]).T / -2.0   # position of the corners in the object frame
        object_pos_rec = mpatches.Rectangle([O_pos_corner_o[0], O_pos_corner_o[1]], self.W_o, self.L_o, 0.0, fc=(1, 1, 1, 0.1), ec=(1, 0, 1, 0.8))
        self.fig_object_pos = self.ax_main.add_artist(object_pos_rec)

        # robot history path 
        self.fig_robot_path = self.ax_main.plot(0.0, 0.0, c='g', ls='-', lw=2.0)

        # the predicted final state
        # W_posn_o_cir = mpatches.Circle(np.array([self.W_posn_o[0],self.W_posn_o[1]]), 0.1, fc=(1, 0.4, 1, 0.8), ec=(0, 0, 1, 0.8), alpha = 0.4)
        # self.fig_W_posn_o = self.ax_main.add_artist(W_posn_o_cir)

        # goal
        W_goal_cir = mpatches.Circle(np.array([self.obj_pos_goal[0],self.obj_pos_goal[1]]), 0.1, fc=(1, 0.1, 1, 0.8), ec=(0, 1, 1, 0.8), alpha = 1.)
        self.fig_W_goal_cir = self.ax_main.add_artist(W_goal_cir)

        plt.draw() 

    def close(self):
        plt.close(self.fig_main)

    def init_path_candidates_both(self, N):
        self.mpc_planned_path = ['path_i']*N
        self.mpc_planned_path_object = ['path_i']*N
        for i in range(N):
            self.mpc_planned_path[i] = self.ax_main.plot(0.0, 0.0, c="red", ls='-', lw=2.0)
            self.mpc_planned_path_object[i] = self.ax_main.plot(0.0, 0.0, c="magenta", ls='-', lw=2.0)

    def show_paths_candidates_robot(self, mpc_path):
        all_x_coord = []
        all_y_coord = []
        all_x_coord_object = []
        all_y_coord_object = []

        for i in range(0, mpc_path.shape[1]):
            x_coord = mpc_path[:, i, 0]  # a list of x in one episode
            y_coord = mpc_path[:, i, 1]  # a list of y in one episode
            all_x_coord.append(x_coord)
            all_y_coord.append(y_coord)

        for j in range(len(self.mpc_planned_path)):
            self.mpc_planned_path[j][0].set_data(np.concatenate((all_x_coord[j], all_y_coord[j])).reshape((2, -1)))

        # Update canvas
        self.fig_main.canvas.draw()
        self.fig_main.canvas.flush_events()


    def save_trajectory(self):
        # action [acc_r, acc_omega]
        # robot state: rob_state [x, y, theta, vr, omega]
        # object state: self.obj_state [x, y, theta, vx, vy, omega]

        full_state = np.append(np.append(np.append(self.prev_rob_state, self.prev_obj_state), self.action), self.time)
        full_state = full_state.reshape(1, -1)
        self.full_trajectory = np.concatenate((self.full_trajectory, full_state), axis=0)

    def save_trajectory_data(self, i_episode, save_directory):
        # save_directory = '/home/susan/Documents/_isaac/sim_yu/trajectories/trajectories_collect/push_trajectory'
        np.save(save_directory + str(i_episode) + '.npy', self.full_trajectory)

    def run_feedback_obs_free_navi(self, robot_pose_current, robot_pos_goal):
        # feedback control law 
        robot_pos_current = robot_pose_current[0:2]
        robot_theta_current = robot_pose_current[2]
        px_error = robot_pos_goal[0] - robot_pos_current[0]
        py_error = robot_pos_goal[1] - robot_pos_current[1]
        rho = np.sqrt(px_error ** 2 + py_error ** 2)
        alpha = -robot_theta_current + np.arctan2(py_error, px_error)
        phi = -robot_theta_current - alpha
        vel = self.pr.k_rho * rho
        omega = self.pr.k_alpha * alpha + self.pr.k_phi * phi
        factor = min(1.0, self.pr.robot_maxVel / vel)
        robot_control_current = [vel * factor, omega * factor]

        return robot_control_current